#ifndef _TRIANGULATE_H
#define _TRIANGULATE_H

#include <math.h>

#define PI 3.1415926536f

typedef struct TTPoint
{
public:
	float x,y;

	TTPoint() {};
	TTPoint(float i, float j) { x=i; y=j; }
	friend TTPoint operator + (const TTPoint& p1, const TTPoint& p2)
		{return TTPoint(p1.x+p2.x, p1.y+p2.y);}
	friend TTPoint operator - (const TTPoint& p1, const TTPoint& p2)
		{return TTPoint(p1.x-p2.x, p1.y-p2.y);}
	friend TTPoint operator * (const TTPoint& p1, float n2)
		{return TTPoint(p1.x*n2, p1.y*n2);}
	friend TTPoint operator * (float n1, const TTPoint& p2)
		{return TTPoint(n1*p2.x, n1*p2.y);}
	friend TTPoint operator / (const TTPoint& p1, float n2)
		{return TTPoint(p1.x/n2, p1.y/n2);}
	friend bool operator == (const TTPoint& p1, const TTPoint& p2)
		{return p1.x==p2.x && p1.y==p2.y;}
	friend bool operator != (const TTPoint& p1, const TTPoint& p2)
		{return p1.x!=p2.x || p1.y!=p2.y;}

	float GetSize(void) const { return (float)sqrt(x*x + y*y); }
} TTPoint;

inline float GetInnerProduct(const TTPoint& pa, const TTPoint& pb)
	{ return pa.x * pb.x + pa.y * pb.y; }

bool IsInTriangle(TTPoint cp, TTPoint p1, TTPoint p2, TTPoint p3);

class Indices
{
public:
	int *index;
	int count;

	Indices();
	Indices(int i);
	virtual ~Indices();
	int Add(int i);
	int Insert(int i, int newindex);
	Indices *Clone(void);
};

class IndicesGroup
{
public:
	Indices **index;
	int count;

	IndicesGroup();
	virtual ~IndicesGroup();
	int Add(Indices *newindex);
	Indices *RemoveTop(void);
};

class TTPolygon
{
protected:
	void Triangulate(IndicesGroup *group);
public:
	TTPoint *pts;
	int ptscount;
	Indices *indices;
	IndicesGroup *triangle;

	TTPolygon();
	virtual ~TTPolygon();
	int Add(TTPoint p);
	int Insert(int index, TTPoint p);
	void Close();
	int AddIndex(int i);
	void InitIndex(void);
	int Triangulate(void);
	void Invert(void);
	bool IsClockwise(void);
};

#endif //_TRIANGULATE_H
